CAST = "cast"
CHROME = "chrome"
CHROMIUM = "chromium"
CONTAINS = "contains"
DISPLAY = "display"
EAGER = "eager"
ENDSWITH = "endswith"
FIREFOX = "firefox"
GET = "get"
HEADERS = "headers"
ICONTAINS = "icontains"
IENDSWITH = "iendswith"
IEXACT = "iexact"
ISTARTSWITH = "istartswith"
ID = "id"
IN = "in"
IIN = "iin"
INPUT_TAG = " <YNK:#> "
NONE = "none"
NORMAL = "normal"
NULL = "null"
QUICK = "quick"
RANK = "rank"
REGEX = "regex"
SESSION = "session"
STARTSWITH = "startswith"
TRANSIENT = "transient"
TYPE = "type"
UNIQUE = "unique"
URL = "url"
WEIGHT = "weight"
YANKED_AT = "yanked_at"
